(* Lightweight thread library for Objective Caml
 * http://www.ocsigen.org/lwt
 * Interface Lwt_sys
 * Copyright (C) 2011 Jérémie Dimino
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, with linking exceptions;
 * either version 2.1 of the License, or (at your option) any later
 * version. See COPYING file for details.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *)

(** System informations. *)

exception Not_available of string
  (** [Not_available(feature)] is an exception that may be raised when
      a feature is not available on the current system. *)

val windows : bool
  (** [true] iff running on windows. *)

(** Features that can be tested. *)
type feature =
    [ `wait4
    | `get_cpu
    | `get_affinity
    | `set_affinity
    | `recv_msg
    | `send_msg
    | `fd_passing
    | `get_credentials
    | `mincore
    | `madvise
    | `fdatasync
    | `libev ]

val have : feature -> bool
  (** Test whether the given feature is available on the current
      system. *)

type byte_order = Little_endian | Big_endian
    (** Type of byte order *)

val byte_order : byte_order
  (** The byte order used by the computer running the program. *)
